# -*- coding: utf-8 -*-
from typing import List, Tuple, Set, Optional

import click
import logging
import os

from babelfish import Error as BabelfishError, Language

from cleanit.config import Config
from cleanit.rule import Rules
from cleanit.subtitle import Subtitle


logger = logging.getLogger('cleanit')


class LanguageParamType(click.ParamType):
    name = 'language'

    def convert(self, value, param, ctx):
        try:
            return Language.fromietf(value)
        except (BabelfishError, ValueError):
            self.fail(f"{click.style(f'{value}', bold=True)} is not a valid language")


LANGUAGE = LanguageParamType()


@click.command()
@click.option('-c', '--config', type=click.Path(), help='YAML config file to be used')
@click.option('-l', '--language', type=LANGUAGE, multiple=True, help='Language as IETF code, '
              'e.g. en, pt-BR (can be used multiple times).')
@click.option('-t', '--tag', required=True, multiple=True, help='Rule tags to be used, '
              'e.g. ocr, tidy, no-sdh, no-style, no-lyrics, no-spam (can be used multiple times). ')
@click.option('-e', '--encoding', help='Save subtitles using the following encoding.')
@click.option('-f', '--force', is_flag=True, default=False,
              help='Force saving the subtitle even if there was no change.')
@click.option('--test', is_flag=True, help='Do not make any change. Useful to be used together with --debug')
@click.option('--debug', is_flag=True, help='Print useful information for debugging and for reporting bugs.')
@click.option('-v', '--verbose', count=True, help='Display debug messages')
@click.argument('path', type=click.Path(), required=True, nargs=-1)
def cleanit(config: Optional[str], language: Optional[Tuple[Language]], tag: Tuple[str],
            encoding: Optional[str],
            force: bool, test: bool, debug: bool, verbose: int, path: Tuple[str]):
    if debug:
        handler = logging.StreamHandler()
        handler.setFormatter(logging.Formatter(logging.BASIC_FORMAT))
        logger.addHandler(handler)
        logger.setLevel(logging.DEBUG)

    languages = set(language)
    tags = set(tag)

    if config and (not os.path.isfile(config) or os.path.isdir(config)):
        click.echo(f"Invalid configuration is defined: {click.style(config, bold=True)}")
        return

    cfg = Config.from_path(config)

    rules = cfg.select_rules(tags=tags, languages=languages)
    if not rules:
        click.echo(f"No rules defined for "
                   f"{click.style(', '.join(tag + tuple(str(lang) for lang in languages)), bold=True)}")
        return

    collected_subtitles: List[Subtitle] = []
    filtered_out_paths: List[str] = []
    discarded_paths: List[str] = []

    if debug or verbose > 1:
        for p in path:
            scan(p, languages, collected_subtitles, filtered_out_paths, discarded_paths)
    else:
        with click.progressbar(path, label='Collecting subtitles', item_show_func=lambda item: item or '') as bar:
            for p in bar:
                scan(p, languages, collected_subtitles, filtered_out_paths, discarded_paths)

    if debug or verbose > 1:
        if verbose > 2:
            for p in filtered_out_paths:
                click.echo(f"{click.style(p, fg='yellow', bold=True)} filtered out")
        for p in discarded_paths:
            click.echo(f"{click.style(p, fg='red', bold=True)} discarded")

    # report collected subtitles
    click.echo(f"{click.style(str(len(collected_subtitles)), bold=True, fg='green')} "
               f"subtitle{'s' if len(collected_subtitles) > 1 else ''} collected / "
               f"{click.style(str(len(filtered_out_paths)), bold=True, fg='yellow')} "
               f"subtitle{'s' if len(filtered_out_paths) > 1 else ''} filtered out / "
               f"{click.style(str(len(discarded_paths)), bold=True, fg='red')} "
               f"path{'s' if len(discarded_paths) > 1 else ''} ignored")

    if not collected_subtitles:
        return

    if debug or verbose > 1:
        for sub in collected_subtitles:
            clean_subtitle(sub, rules, encoding, force, test, verbose)
    else:
        saved_count = 0
        with click.progressbar(collected_subtitles, label='Cleaning subtitles',
                               item_show_func=lambda s: click.style(s and s.name or '', bold=True)) as bar:
            for sub in bar:
                saved = clean_subtitle(sub, rules, encoding, force, test, verbose)
                saved_count += 1 if saved else 0

        # report processed subtitles
        unchanged_count = len(collected_subtitles) - saved_count
        click.echo(f"{click.style(str(saved_count), bold=True, fg='green')} "
                   f"subtitle{'s' if saved_count > 1 else ''} saved / "
                   f"{click.style(str(unchanged_count), bold=True, fg='yellow')} "
                   f"subtitle{'s' if unchanged_count > 1 else ''} unchanged")


def clean_subtitle(sub: Subtitle, rules: Rules, encoding: Optional[str], force: bool, test: bool, verbose: int,
                   echo=False):
    try:
        modified = sub.clean(rules)
        if (modified or force) and not test:
            sub.save(encoding=encoding)
            if echo:
                click.echo(f"{click.style(sub.name, fg='green', bold=True)} saved.")
            return True
        elif echo and verbose > 1:
            click.echo(f"No modifications for {click.style(sub.name, fg='yellow')}", )

        return False
    finally:
        # to free up memory
        sub.finalize()


def scan(path: str, languages: Set[Language], collected: List[Subtitle], filtered_out: List[str], discarded: List[str]):
    if not os.path.exists(path):
        discarded.append(path)

    elif os.path.isfile(path):
        if path.lower().endswith('.srt'):
            subtitle = Subtitle(path)
            if subtitle.match(languages):
                collected.append(subtitle)
            else:
                filtered_out.append(path)

    elif os.path.isdir(path):
        for dir_path, dir_names, file_names in os.walk(path):
            for filename in file_names:
                file_path = os.path.join(dir_path, filename)
                scan(file_path, languages, collected, filtered_out, discarded)



# -*- coding: utf-8 -*-
from typing import List


def string_enum(values: List[str]):
    return {'type': 'string', 'enum': values}


def array(schema: dict, min_items=0):
    return {'type': 'array', 'items': schema, 'minItems': min_items}


def one_of(*schemas: dict):
    return {'oneOf': list(schemas)}


def one_or_array(schema: dict):
    return one_of(schema, array(schema, min_items=1))


class TagsSchema:
    schema = {'type': 'string'}


class RuleTypeSchema:
    schema = string_enum(['text', 'regex'])


class MatchTypeSchema:
    schema = string_enum(['exact', 'contains', 'startswith', 'endswith'])


class FlagSchema:
    schema = string_enum(['ignorecase', 'dotall', 'multiline', 'locale', 'unicode', 'verbose'])


class ReplacementSchema:
    schema = {'type': 'string'}


class PrioritySchema:
    schema = {'type': 'integer'}


class LanguageSchema:
    schema = {'type': 'string'}


class AliasesSchema:
    schema = {
        'type': 'object',
        'patternProperties': {
            '.+': {'type': 'string'}
        },
        'minProperties': 1
    }


class ExampleSchema:
    schema = one_of(array({'type': 'string'}, min_items=1),
                    {'type': 'object',
                     'patternProperties': {
                         '.+': {'type': ['string', 'null']}
                     },
                     'minProperties': 1
                     })


class RuleSchema:
    schema = {
        'type': 'object',
        'properties': {
            'type': RuleTypeSchema.schema,
            'match': MatchTypeSchema.schema,
            'flags': one_or_array(FlagSchema.schema),
            'tags': one_or_array(TagsSchema.schema),
            'priority': PrioritySchema.schema,
            'languages': one_or_array(LanguageSchema.schema),
            'patterns': one_or_array({'type': 'string'}),
            'disabled': {'type': ['boolean', 'null']},
            'replacement': ReplacementSchema.schema,
            'examples': ExampleSchema.schema
        },
        'additionalProperties': False
    }


class RulesSchema:
    schema = {
        'type': 'object',
        'patternProperties': {
            '.+': RuleSchema.schema
        },
        'minProperties': 1
    }


class RootSchema:
    schema = {
        'type': 'object',
        'properties': {
            'aliases': AliasesSchema.schema,
            'templates': array(RuleSchema.schema, min_items=0),
            'rules': RulesSchema.schema
        },
        'additionalProperties': False
    }



# -*- coding: utf-8 -*-
import os
import logging
import re
from collections import UserList
from typing import Optional, Collection, Dict, Tuple, Set, Iterable, List

from babelfish import Language
from pysrt import SubRipItem, SubRipFile

from .utils import ensure_list, get_language_groups

logger = logging.getLogger(__name__)


class Change:
    def __init__(self, item: SubRipItem):
        self._max_chars = 50
        self.start = item.start
        self.end = item.end
        self.a_lines = [line for line in item.text.split('\n')]
        self.b_lines = self.a_lines
        self.rules: List[str] = []

    def track(self, rule: str, text: Optional[str]):
        self.rules.append(rule)
        self.b_lines = [line for line in text.split('\n')] if text else []

    @property
    def timestamp(self):
        return f'{self.start} --> {self.end}'

    @property
    def rules_description(self):
        return ' '.join(r for r in self.rules)

    @property
    def max_chars(self):
        return max(self._max_chars, (len(self.rules_description) // 2 + 1),
                   *[len(line) for line in self.a_lines + self.b_lines])

    @max_chars.setter
    def max_chars(self, value: int):
        self._max_chars = max(self._max_chars, value)

    def get_line_tuples(self):
        a_lines = self.a_lines + [''] * (len(self.b_lines) - len(self.a_lines))
        b_lines = self.b_lines + [''] * (len(self.a_lines) - len(self.b_lines))
        for a, b in zip(a_lines, b_lines):
            yield a, b

    def __repr__(self):
        return f'<{self.__class__.__name__} [{self.timestamp} ({self.rules_description})]>'

    def __str__(self):
        max_width = self.max_chars * 2 + 1
        lines = [
            f'<{self.rules_description}>'.center(max_width),
            f'{self.timestamp.center(max_width//2)} {self.timestamp.center(max_width//2)}'
        ]
        for a, b in self.get_line_tuples():
            line = f'{a.center(max_width//2)}|{b.center(max_width//2)}'
            lines.append(line)

        return '\n'.join(lines)


class Changes(UserList):

    def __init__(self, srt: SubRipFile):
        super().__init__()
        self.path = srt.path

    def __repr__(self):
        return f'<{self.__class__.__name__} [{self.path}:{len(self)}]>'

    def __str__(self):
        lines = [f'{self.path}']
        for item in self:
            lines.append(str(item))

        return '\n\n'.join(lines) + '\n'


class Rule:
    def __init__(self, name: str, aliases: Dict[str, str], patterns: Collection[str] or str, type: str = 'regex',
                 match: str = 'contains', priority: int = 0, tags: Optional[Collection[str] or str] = None,
                 flags: Optional[Collection[str] or str] = None, languages: Optional[Collection[str] or str] = None,
                 replacement: Optional[str] = None, examples: Optional[Dict[str, str] or str] = None, disabled=False):
        self.name = name
        re_flags = 0
        r_flags = set(ensure_list(flags))
        for flag in r_flags:
            re_flags |= re.__dict__.get(flag.upper())

        regexes = []
        for pattern in ensure_list(patterns):
            re_pattern: str = pattern if type == 'regex' else re.escape(pattern)
            if type == 'regex':
                for k, v in aliases.items():
                    re_pattern = re_pattern.replace(k, v)

            if match in ('endswith', 'exact'):
                re_pattern = f'{re_pattern}$'
            if match in ('startswith', 'exact'):
                re_pattern = f'^{re_pattern}'
            regexes.append(re.compile(re_pattern, re_flags))

        self.regexes = regexes
        self.flags = r_flags
        self.priority = priority
        self.tags = set(ensure_list(tags))
        self.disabled = disabled
        self.languages = {Language.fromietf(lang) for lang in ensure_list(languages)}
        self.language_groups = get_language_groups(self.languages)
        self.replacement = replacement
        self.examples = {key.strip(): value.strip() if value is not None else None for key, value in
                         (examples if isinstance(examples, dict) else (
                             {e: None for e in examples} if isinstance(examples, list) else {})).items()}

    def __repr__(self):
        return f'<{self.__class__.__name__} [{str(self)}]>'

    def __str__(self):
        return self.name

    def matches(self, tags: Set[str] = None, languages: Set[Language] = None):
        if self.disabled:
            return False

        tags = tags or set()
        if self.tags and not self.tags.intersection(tags) and 'all' not in tags:
            return False

        if not self.languages or not languages:
            return True

        language_groups = get_language_groups(languages)

        return self.language_groups.intersection(language_groups)

    def apply(self, text: str or None, replacement: str = None, change: Change = None):
        for regex in self.regexes:
            m = regex.search(text)
            if not m:
                continue

            repl = replacement or self.replacement
            if repl is None:
                if change:
                    change.track(self.name, None)
                return None, True

            value = regex.sub(repl, text).strip() or None
            if change:
                change.track(self.name, value or None)

            if not value:
                return None, True

            return self.apply(value, replacement=replacement, change=change)[0], True

        return text, False


class Rules(UserList):

    def __init__(self, rules: Iterable[Rule], tags: Set[str] = None, languages: Set[Language] = None):
        self.tags = tags or set()
        self.languages = languages or set()
        super().__init__([r for r in rules if r.matches(tags=self.tags, languages=self.languages)])

    def __repr__(self):
        return f'<{self.__class__.__name__} [{self.languages}:{self.tags}:{len(self)}]>'

    def apply(self, value: str, replacement: str = None, change: Change = None) -> Tuple[Optional[str], bool]:
        result = value
        modified = False
        for rule in self:
            previous = result
            result, matches = rule.apply(previous, replacement=replacement, change=change)
            if matches:
                modified = True
                if not result:
                    return result, modified
                else:
                    break

        if modified:
            return self.apply(result, replacement=replacement, change=change)[0], True

        return result, modified

    def clean(self, value: str, replacement=''):
        return self.apply(value, replacement=replacement)[0]



import copy
import json
import pkgutil
from types import GeneratorType
from typing import Any, Set

import jsonschema
import yaml
from babelfish import Language

from .schema import RootSchema


def is_iterable(obj: Any):
    return hasattr(obj, '__iter__') and not isinstance(obj, str) or isinstance(obj, GeneratorType)


def ensure_list(param: Any):
    if not param:
        param = []
    elif not is_iterable(param):
        param = [param]
    return param


def get_language_groups(languages: Set[Language]):
    groups = set()
    for language in languages:
        ietf = str(language)
        ietf_parts = ietf.split('-')

        groups.add(ietf)
        groups.add(ietf_parts[0])
        groups.add('-'.join(ietf_parts[:2]))

    return groups


def validate(data: dict):
    jsonschema.validate(data, RootSchema.schema)


def load_config_file(path: str):
    with open(path, 'r') as f:
        data = json.load(f) if path.endswith('.json') else yaml.safe_load(f.read())
    validate(data)
    return data


def load_config_resource(resource_name: str):
    resource_data = pkgutil.get_data('cleanit', resource_name)
    data = json.loads(resource_data) if resource_name.endswith('.json') else yaml.safe_load(resource_data)
    validate(data)
    return data


def merge_options(*options):
    """
    Merge options into a single options dict.
    :param options:
    :type options:
    :return:
    :rtype:
    """

    merged = {}
    if options:
        if options[0]:
            merged.update(copy.deepcopy(options[0]))

        for options in options[1:]:
            if options:
                pristine = options.get('pristine')

                if pristine is True:
                    merged = {}
                elif pristine:
                    for to_reset in pristine:
                        if to_reset in merged:
                            del merged[to_reset]

                for (option, value) in options.items():
                    merge_option_value(option, value, merged)

    return merged


def merge_option_value(option, value, merged):
    """
    Merge option value
    :param option:
    :param value:
    :param merged:
    :return:
    """
    if value is not None and option != 'pristine':
        if option in merged.keys() and isinstance(merged[option], list):
            for val in value:
                if val not in merged[option] and val is not None:
                    merged[option].append(val)
        elif option in merged.keys() and isinstance(merged[option], dict):
            merged[option] = merge_options(merged[option], value)
        elif isinstance(value, list):
            merged[option] = list(value)
        else:
            merged[option] = value



# -*- coding: utf-8 -*-

default_config_resources = [
    'cleanit.yml',
    'cleanit-ocr.en.yml',
    'cleanit-no-spam.en.yml',
    'cleanit-tidy.en.yml',
    'cleanit-ocr.pt.yml',
    'cleanit-tidy.pt.yml',
    'cleanit-no-spam.pt.yml',
    'cleanit-ocr.yml',
    'cleanit-no-style.yml',
    'cleanit-no-sdh.yml',
    'cleanit-no-lyrics.yml',
    'cleanit-no-spam.yml',
    'cleanit-tidy.yml',
]



# -*- coding: utf-8 -*-
import logging
import os
from io import StringIO
from typing import Optional, Set

import chardet
import pysrt
from babelfish import language_converters, country_converters, Language, LanguageReverseConverter, LanguageReverseError
from pysrt import SubRipFile

from .rule import Rules, Change, Changes

logger = logging.getLogger(__name__)


class CleanitLanguageConverter(LanguageReverseConverter):
    @property
    def codes(self):
        return (language_converters['alpha3b'].codes |
                language_converters['alpha2'].codes |
                language_converters['name'].codes |
                language_converters['opensubtitles'].codes |
                country_converters['name'].codes)

    def convert(self, alpha3: str, country: str = None, script: str = None):
        return str(Language(alpha3, country, script))

    def reverse(self, name: str):
        name = name.lower()
        for conv in [Language.fromietf,
                     Language,
                     Language.fromalpha3b,
                     Language.fromalpha2,
                     Language.fromname,
                     Language.fromopensubtitles]:
            try:
                reverse = conv(name)
                return reverse.alpha3, reverse.country, reverse.script
            except (ValueError, LanguageReverseError):
                pass

        return 'und', None, None


language_converters['cleanit'] = CleanitLanguageConverter()


def get_subtitle_language(path: str):
    lang = os.path.splitext(os.path.splitext(path)[0])[1] or '.und'
    return Language.fromcleanit(lang[1:])


class Subtitle:
    def __init__(self, path: str, encoding: str = None):
        self.path = path
        self.language = get_subtitle_language(path)
        self.encoding = encoding
        self.subtitle: Optional[SubRipFile] = None

    def __repr__(self):
        return f'<{self.__class__.__name__} [{self.path}]>'

    @property
    def content(self):
        if self.subtitle:
            writer = StringIO()
            self.subtitle.write_into(writer)
            writer.seek(0)
            return writer.read().strip()

    @property
    def name(self):
        return os.path.split(self.path)[1]

    def match(self, languages: Set[Language]):
        return not languages or self.language in languages

    def save(self, path: str = None, encoding: str = None):
        self.subtitle.save(path=path or self.path, encoding=encoding or self.encoding)

    def guess_encoding(self):
        with open(self.path, 'rb') as f:
            content = f.read()

        # always try utf-8 first
        encodings = ['utf-8']

        # add language-specific encodings
        if self.language.alpha3 == 'zho':
            encodings.extend(['gb18030', 'big5'])
        elif self.language.alpha3 == 'jpn':
            encodings.append('shift-jis')
        elif self.language.alpha3 == 'ara':
            encodings.append('windows-1256')
        elif self.language.alpha3 == 'heb':
            encodings.append('windows-1255')
        elif self.language.alpha3 == 'tur':
            encodings.extend(['iso-8859-9', 'windows-1254'])
        elif self.language.alpha3 == 'pol':
            # Eastern European Group 1
            encodings.extend(['windows-1250'])
        elif self.language.alpha3 == 'bul':
            # Eastern European Group 2
            encodings.extend(['windows-1251'])
        else:
            # Western European (windows-1252)
            encodings.append('latin-1')

        # try to decode
        for encoding in encodings:
            try:
                content.decode(encoding)
            except UnicodeDecodeError:
                pass
            else:
                return encoding

        logger.warning('Could not guess encoding from language')

        # fallback on chardet
        encoding = chardet.detect(content)['encoding']
        logger.info(f'Chardet found encoding {encoding}')

        return encoding

    def clean(self, rules: Rules, clean_indexes=True):
        rules = Rules(rules=rules, tags=rules.tags, languages={self.language})
        self.encoding = self.encoding or self.guess_encoding()
        self.subtitle = pysrt.open(self.path, encoding=self.encoding)
        track_changes = logger.isEnabledFor(logging.DEBUG)
        changes = Changes(self.subtitle) if track_changes else None

        modified = False
        for i, item in reversed(list(enumerate(self.subtitle))):
            change = Change(item) if track_changes else None
            text, changed = rules.apply(item.text, change=change)
            if changed:
                modified = True
                if not text:
                    del self.subtitle[i]
                else:
                    item.text = text
                if track_changes:
                    changes.append(change)

        if modified:
            if track_changes:
                max_chars = max([c.max_chars for c in changes])
                for c in changes:
                    c.max_chars = max_chars

                logger.debug(f'Changes for {changes}')
            if clean_indexes:
                self.subtitle.clean_indexes()

        return modified

    def finalize(self):
        self.subtitle = None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import io
import os
import re

from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
data_path = os.path.join(here, 'cleanit', 'data')

locations = [location for location in os.listdir(data_path) if location.endswith(('.yml', '.yaml', '.json'))]
with open(os.path.join(data_path, '__init__.py'), 'w') as f:
    f.write('# -*- coding: utf-8 -*-\n\n')
    f.write('default_config_resources = [\n')
    for location in locations:
        f.write(f"    '{location}',\n")
    f.write(']\n')


def read(*parts):
    # intentionally *not* adding an encoding option to open, See:
    #   https://github.com/pypa/virtualenv/issues/201#issuecomment-3145690
    return io.open(os.path.join(here, *parts), 'r').read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r'^__version__ = [\'"]([^\'"]*)[\'"]', version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError('Unable to find version string.')


# requirements
install_requires = [
    'appdirs>=1.4.4',
    'babelfish>=0.5.5',
    'chardet>=4.0.0',
    'click>=7.1.2',
    'jsonschema>=3.2.0',
    'pysrt>=1.1.2',
    'pyyaml>=5.4.1'
]
tests_require = [
    'pytest>=6.2.2',
    'pytest-cov>=2.11.1',
    'pytest-flake8>=1.0.7'
]
dev_requirements = [
    'sphinx'
]
package_data = [
    'data/*'
]

entry_points = {
    'console_scripts': [
        'cleanit = cleanit.cli:cleanit'
    ]
}

setup(name='cleanit',
      version=find_version('cleanit', '__init__.py'),
      license='Apache License 2.0',
      description='Subtitles extremely clean',
      long_description='{readme}\n\n{history}'.format(readme=read('README.rst'), history=read('HISTORY.rst')),
      long_description_content_type='text/markdown',
      keywords='subtitle subtitles clean blacklist replace ocr fix tidy',
      url='https://github.com/ratoaq2/cleanit',
      author='Rato',
      author_email='ratoaq2@gmail.com',
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: Apache Software License',
          'Operating System :: OS Independent',
          'Programming Language :: Python',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Programming Language :: Python :: 3.9',
          'Topic :: Software Development :: Libraries :: Python Modules',
          'Topic :: Multimedia :: Video'
      ],
      packages=find_packages(),
      package_dir={'cleanit': 'cleanit'},
      package_data={'cleanit': package_data},
      # include_package_data=True,
      zip_safe=True,
      entry_points=entry_points,
      install_requires=install_requires,
      tests_require=tests_require,
      extras_require={
          'tests': tests_require,
          'dev': dev_requirements
      })



# -*- coding: utf-8 -*-
import os
import re
from typing import List, Collection

import logging

from . import __title__, __author__
from .data import default_config_resources
from .rule import Rule, Rules
from .utils import load_config_file, merge_options, load_config_resource

from appdirs import AppDirs
from babelfish import Language

logger = logging.getLogger(__name__)
here = os.path.abspath(os.path.dirname(__file__))

FILENAME_RE = re.compile(r'^cleanit(-[\w-]+)?(\.\w+(-\w+){0,2})?.(ya?ml|json)$')


def get_default_config_files():
    dirs = AppDirs(appname=__title__, appauthor=__author__)
    folders = [dirs.site_config_dir, dirs.user_config_dir]

    locations = []
    for folder in folders:
        try:
            locations.extend([os.path.abspath(os.path.join(folder, f)) for f in os.listdir(folder) if FILENAME_RE.search(f)])
        except FileNotFoundError as e:
            logger.debug(f"Discarding location '{folder}'. {str(e)}")

    return locations


def load_configuration(*locations: str):
    configurations = []
    for location in locations:
        data = load_config_file(location)
        configurations.append(data)
        logger.debug(f'Loaded configuration from {location}')
    return configurations


def load_default_resources():
    configurations = []
    for resource in default_config_resources:
        data = load_config_resource(f'data/{resource}')
        configurations.append(data)
        logger.debug(f'Loaded configuration from {resource}')
    return configurations


default_config = merge_options(*(load_default_resources() + load_configuration(*get_default_config_files())))


class Config:
    def __init__(self, *args: dict):
        self.data = merge_options(default_config, *args)
        aliases = self.data.get('aliases', {})
        rules: List[Rule] = []

        for name, r in self.data.get('rules', {}).items():
            rules.append(Rule(name=name, aliases=aliases, **r))

        rules.sort(key=lambda s: s.priority, reverse=True)
        self.rules = rules

    @classmethod
    def from_path(cls, path: str = None):
        folders = [path] if path and os.path.isdir(path) else []
        locations = [path] if path and os.path.isfile(path) else []

        for folder in folders:
            locations.extend([f for f in os.listdir(folder) if FILENAME_RE.search(f)])

        return Config(*load_configuration(*locations))

    def select_rules(self, tags: Collection[str] = None, languages: Collection[Language] = None):
        return Rules(rules=self.rules, tags=tags, languages=languages)
